#  Copyright (c) 2022 Szymon Mikler

from __future__ import annotations

import torch
from torch import nn


def get_parameter_count(model: nn.Module):
    """Get the number of parameters of a model."""
    cnt = 0
    for param in model.parameters():
        cnt += param.shape.numel()
    return cnt


def get_parameter_shapes(model: nn.Module):
    """Get the shapes of parameters of a model."""
    shapes = []
    for param in model.parameters():
        shapes.append(tuple(param.shape))
    return shapes


def model_similar(a: nn.Module, b: nn.Module):
    """Check whether two models have the same number of parameters and the same shapes of parameters."""
    if get_parameter_count(a) != get_parameter_count(b):
        return False

    if sorted(get_parameter_shapes(a)) != sorted(get_parameter_shapes(b)):
        return False
    return True


def hash_torch_tensor(tensor: torch.Tensor):
    """Interpret the tensor as a string and return its hash."""
    tensor_as_string = str(tensor.flatten().tolist())
    return hash(tensor_as_string)


def models_have_corresponding_parameters(a: nn.Module, b: nn.Module):
    """Check whether two models' parameters have identical hash values.

    Parameter order does not matter.
    So if two models have identical parameters but in different order, this will still return True.
    """
    hashes_a = [hash_torch_tensor(p) for p in a.parameters()]
    hashes_b = [hash_torch_tensor(p) for p in b.parameters()]
    return set(hashes_a) == set(hashes_b)



#  Copyright (c) 2022 Szymon Mikler

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Set, Tuple

from .symbolic_tensor import SymbolicTensor

if TYPE_CHECKING:
    from .symbolic_model import SymbolicModel

from collections import defaultdict

from torch import nn


def check_for_missing_inputs(
    used_nodes: Set[SymbolicTensor],
    inputs: Tuple[SymbolicTensor, ...] | List[SymbolicTensor],
):
    """Check if there exist nodes which require input from outside of the graph.

    It is forbidden, as it doesn't make sense.

    Example of such violation::

        x1 = Input(shape=(32,))
        x2 = Input(shape=(32,))
        x3 = x1 + x2
        model = SymbolicModel(inputs=x1, outputs=x3)

    Model cannot execute defined operations unless ``x2`` is given, because ``x3`` requires it.
    """
    for node in used_nodes:
        if node in inputs:
            continue
        for parent in node.parents:
            assert parent in used_nodes, (
                f"Node {node} depends on the output of a foreign node! " f"Perhaps you set the wrong inputs?"
            )


def figure_out_nodes_between(
    inputs: Tuple[SymbolicTensor, ...] | List[SymbolicTensor] | None = None,
    outputs: Tuple[SymbolicTensor, ...] | List[SymbolicTensor] | None = None,
) -> Set[SymbolicTensor]:
    """Returns intersection of predecessors tree of outputs and succesors tree of inputs."""

    all_nodes_above: List[SymbolicTensor] = []
    if outputs is not None:
        for output_leaf in outputs:
            nodes_above = output_leaf._get_all_nodes_above()
            all_nodes_above.extend(nodes_above)

    all_nodes_below: List[SymbolicTensor] = []
    if inputs is not None:
        for input_leaf in inputs:
            nodes_below = input_leaf._get_all_nodes_below()
            all_nodes_below.extend(nodes_below)

    # Get the intersection
    if inputs is None or outputs is None:
        used_nodes = set(all_nodes_above) | set(all_nodes_below)
    else:
        used_nodes = set(all_nodes_above) & set(all_nodes_below)
    if inputs is not None:
        check_for_missing_inputs(used_nodes, inputs)
    return used_nodes


def default_node_text(plh: SymbolicTensor) -> str:
    return str(plh.shape)


def default_edge_text(layer: nn.Module) -> str:
    return layer._get_name()


def _calc_sum_sq_distances(graph, positions_dict):
    sum_sq_distances = 0
    for node in positions_dict:
        related_nodes = list(graph.predecessors(node))

        for n2 in related_nodes:
            x1, y1 = positions_dict[node]
            x2, y2 = positions_dict[n2]
            sum_sq_distances += (x1 - x2) ** 2 + (y1 - y2) ** 2
    return sum_sq_distances


def _transpose_positions_dict(pos_dict):
    return {k: tuple(reversed(v)) for k, v in pos_dict.items()}


def _fix_positions_in_multipartite_layout(graph, orig_positions_dict, align: str = "vertical"):
    import networkx as nx
    import numpy as np
    from scipy.optimize import linear_sum_assignment

    assert isinstance(graph, nx.DiGraph)

    if align == "vertical":
        orig_positions_dict = _transpose_positions_dict(orig_positions_dict)

    orig_sum = _calc_sum_sq_distances(graph, orig_positions_dict)

    positions = list(orig_positions_dict.values())
    selected_positions: Dict[int, Tuple[float, float]] = {}

    layers = defaultdict(list)
    for pos in positions:
        layers[pos[1]].append(pos[0])

    layer_to_nodes = defaultdict(list)
    for node, pos in orig_positions_dict.items():
        layer_to_nodes[pos[1]].append(node)

    # Layers are just unique Y positions on the plot
    # Go through the highest layer to the lowest one
    for layer, nodes_in_layer in reversed(sorted(layer_to_nodes.items())):
        n = len(nodes_in_layer)
        distances = np.zeros(shape=(n, n))

        # Extract available X positions in the layer
        avail_xs = layers[layer]

        for node_idx, node in enumerate(nodes_in_layer):
            related_nodes = list(graph.predecessors(node))
            related_nodes += list(graph.successors(node))

            for x_idx, x in enumerate(avail_xs):
                sum_distances = 0
                for related_node in related_nodes:
                    if related_node in selected_positions:
                        r_x, r_y = selected_positions[related_node]
                        sum_distances += (x - r_x) ** 2 + (layer - r_y) ** 2
                    else:
                        # If node doesn't have a position yet
                        # we use the original position but we weight it down
                        r_x, r_y = orig_positions_dict[related_node]
                        sum_distances += ((x - r_x) ** 2 + (layer - r_y) ** 2) / 2
                distances[node_idx, x_idx] = sum_distances

        # Minimize the sum of squared distances
        # This might work poorly when there are interconnections in the layer
        rows, cols = linear_sum_assignment(distances)
        for row, col in zip(rows, cols):
            selected_positions[nodes_in_layer[row]] = (avail_xs[col], layer)

    new_sum = _calc_sum_sq_distances(graph, selected_positions)

    if new_sum <= orig_sum:
        r = selected_positions
    else:
        r = orig_positions_dict

    if align == "vertical":
        r = _transpose_positions_dict(r)

    return r


def variable_name_resolver(namespace):
    def resolver(x):
        names = [name for name in namespace if namespace[name] is x]
        return names[0] if names else "?"

    return resolver


def draw_graph(
    *,
    model: SymbolicModel | None = None,
    inputs: Iterable[SymbolicTensor] | SymbolicTensor | None = None,
    outputs: Iterable[SymbolicTensor] | SymbolicTensor | None = None,
    node_text_func: Callable[[SymbolicTensor], str] | None = None,
    edge_text_func: Callable[[nn.Module], str] | None = None,
    node_text_namespace: Dict[str, Any] | None = None,
    rotate_graph: bool = False,
    rotate_labels: bool = False,
) -> None:
    """Plot graph of the computations, nodes being placeholder variables and nn.Modules being edges.

    This is not suitable for large graphs or large neural networks. This is a simple tool that
    was designed to demonstrate that Pytorch Symbolic creates sensible graphs that are nice
    to visualize.

    Parameters
    ----------
    model
        A SymbolicModel to be plotted. This or ``inputs`` must be provided.
    inputs
        Input in the graph of SymbolicTensor computations. This or ``model`` must be provided.
    node_text_func
        A function that returns text that will be written on Nodes.
    edge_text_func
        A function that returns text that will be written on Edges.
    node_text_namespace
        If used with ``globals()``, it will try to show variable name on each node.
        Has an effect only if `node_text_func` is None.
    rotate_labels
        If True, text on edges will be rotated in the direction of the arrow.
    rotate_graph
        If True, the consecutive layers will be shown to the right, instead of downwards.
    """
    try:
        import matplotlib.patches
        import matplotlib.pyplot as plt
        import networkx as nx
    except ImportError as e:
        print(
            "To plot graphs, you need to install networkx, matplotlib and scipy. Run `pip install networkx`."
        )
        raise e

    from .symbolic_model import SymbolicModel

    if node_text_func is None:
        if node_text_namespace is not None:
            node_text_func = variable_name_resolver(node_text_namespace)
        else:
            node_text_func = default_node_text

    if edge_text_func is None:
        edge_text_func = default_edge_text

    if model is not None:
        assert isinstance(model, SymbolicModel)
        inputs = model.inputs
        outputs = model.outputs

    elif inputs is not None or outputs is not None:
        if inputs is not None:
            if isinstance(inputs, SymbolicTensor):
                inputs = (inputs,)
            assert all(isinstance(x, SymbolicTensor) for x in inputs)
            inputs = tuple(inputs)
        if outputs is not None:
            if isinstance(outputs, SymbolicTensor):
                outputs = (outputs,)
            assert all(isinstance(x, SymbolicTensor) for x in outputs)
            outputs = tuple(outputs)

    else:
        raise KeyError("Provide either `model` or `inputs` or/and `outputs`!")

    used_nodes = figure_out_nodes_between(inputs, outputs)

    graph = nx.DiGraph()
    assert isinstance(inputs, Iterable)
    to_visit = list(inputs)
    visited = set(inputs)

    node_labels = {}
    node_colors = {}
    edge_labels = {}

    INPUT_COLOR = (1.0, 0.3, 0.3)
    OUTPUT_COLOR = (0.3, 1.0, 0.3)
    OTHER_COLOR = (0.0, 1.0, 1.0)

    while to_visit:
        node = to_visit.pop()
        graph.add_node(node, depth=node.depth)

        if inputs and node in inputs:
            node_colors[node] = INPUT_COLOR
        elif outputs and node in outputs:
            node_colors[node] = OUTPUT_COLOR
        else:
            node_colors[node] = OTHER_COLOR

        node_labels[node] = node_text_func(node)

        for parent in node.parents:
            if parent not in used_nodes:
                continue
            assert node.layer is not None
            graph.add_edge(parent, node, layer=node.layer)
            edge_labels[parent, node] = edge_text_func(node.layer)

        for child in node.children:
            if child in used_nodes and child not in visited:
                to_visit.append(child)
                visited.add(child)

    scale = 1 if rotate_graph else -1
    align = "vertical" if rotate_graph else "horizontal"
    pos = nx.multipartite_layout(graph, subset_key="depth", align=align, scale=scale)
    pos = _fix_positions_in_multipartite_layout(graph, pos, align=align)

    nx.draw_networkx(
        graph,
        pos,
        with_labels=True,
        labels=node_labels,
        node_size=1000,
        arrowsize=20,
        node_shape="o",
        node_color=[node_colors[n] for n in graph.nodes],
        font_weight="bold",
        edge_color="grey",
    )

    nx.draw_networkx_edge_labels(
        graph,
        pos,
        edge_labels,
        label_pos=0.5,
        rotate=rotate_labels,
        clip_on=False,
    )

    handles = [
        matplotlib.patches.Patch(color=INPUT_COLOR, label="Input node"),
        matplotlib.patches.Patch(color=OUTPUT_COLOR, label="Output node"),
        matplotlib.patches.Patch(color=OTHER_COLOR, label="Hidden node"),
    ]
    plt.legend(handles=handles)


def topological_sort(nodes: Set[SymbolicTensor]) -> List[SymbolicTensor]:
    """In graph theory, a topological sort or topological ordering of a directed acyclic graph (DAG) is a
    linear ordering of its nodes in which each node comes before all nodes to which it has outbound edges.
    Every DAG has one or more topological sorts."""

    children = {node: set((c for c in node.children if c in nodes)) for node in nodes}
    parents = {node: set((p for p in node.parents if p in nodes)) for node in nodes}
    grandparents = [node for node in nodes if len(parents[node]) == 0]  # no incoming edges

    n_edges = sum(len(v) for v in children.values()) + sum(len(v) for v in parents.values())

    topologically_sorted = []

    while grandparents:
        node = grandparents.pop()
        topologically_sorted.append(node)

        for neighbor in tuple(children[node]):
            children[node].remove(neighbor)
            parents[neighbor].remove(node)

            if len(parents[neighbor]) == 0:
                grandparents.append(neighbor)

    n_edges = sum(len(v) for v in children.values()) + sum(len(v) for v in parents.values())

    assert n_edges == 0, "Graph is not a DAG (directed acyclic graph)!"
    return topologically_sorted



#  Copyright (c) 2022 Szymon Mikler

import torch
from torch import nn


class AnyOpLayer(nn.Module):
    def __init__(self, op):
        super().__init__()
        self.forward = op


class NamedAnyOpLayer(nn.Module):
    def __init__(self, op, name):
        super().__init__()
        self.forward = op
        self.name = name

    def __repr__(self):
        return self.name

    def _get_name(self):
        return self.name


class CallbackLayer(nn.Module):
    def __init__(self, op):
        """This can be used for debugging or logging purposes. Accepts only one argument.

        Example::

            x = CallbackLayer(print)(x)

        It does not change anything in x, but prints its value.
        """
        super().__init__()
        self.callback = op

    def forward(self, arg):
        self.callback(arg)
        return arg


class AddOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a + b


class SubOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a - b


class MulOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a * b


class ModOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return torch.remainder(a, b)


class MatmulOpLayer(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    def forward(a, b):
        return a @ b


class ConcatLayer(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, *tensors):
        return torch.cat(tensors=tensors, dim=self.dim)


class StackLayer(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.dim = dim

    def forward(self, *tensors):
        return torch.stack(tensors=tensors, dim=self.dim)


class ReshapeLayer(nn.Module):
    def __init__(self, shape, batch_size_included=False):
        super().__init__()
        if not batch_size_included:
            self.shape = (-1, *shape)
        else:
            self.shape = shape

    def forward(self, inputs):
        return torch.reshape(input=inputs, shape=self.shape)



#  Copyright (c) 2022 Szymon Mikler

import logging

from .experimental_api import enable_symbolic_API_for_new_modules, remove_call_wrapper_from_all_modules

enable_symbolic_API_for_new_modules()

# Constants
CODEGEN_MIN_LOOP_LENGTH = 50


def optimize_module_calls():
    msg = "Optimizing module calls for existing layers! Reusing them might throw errors!"
    logging.warning(msg)
    remove_call_wrapper_from_all_modules()



#  Copyright (c) 2022 Szymon Mikler

from __future__ import annotations

from typing import Callable, Dict, Hashable, List, Tuple

from torch import nn

from . import useful_layers
from .symbolic_model import SymbolicTensor


def add_module_to_model(module, *args):
    assert isinstance(args[0], SymbolicTensor)
    return args[0](module, *args[1:])


def _replace_symbolic_with_value(container, extracted, navigation):
    """Search recursively for all occurences of SymbolicTensor and replace them with their value.

    At the same time save navigation to know, how to do indexing to get to them.

    If navigation for container ends up as [..., [1, 2, 0, "TEST", 5], ...]
    this means that to get the element you should index container[1][2][0]["TEST"][5].
    """
    if isinstance(container, SymbolicTensor):
        navigation.append(navigation[-1].copy())
        extracted.append(container)
        return container.v

    if isinstance(container, List) or isinstance(container, Tuple):
        new_list = []
        for i, x in enumerate(container):
            navigation[-1].append(i)
            new_list.append(_replace_symbolic_with_value(x, extracted, navigation))
            navigation[-1].pop()
        return new_list
    if isinstance(container, Dict):
        new_dict = {}
        for k, v in container.items():
            navigation[-1].append(k)
            new_dict[k] = _replace_symbolic_with_value(v, extracted, navigation)
            navigation[-1].pop()
        return new_dict
    return container


def add_to_model(func: Callable | nn.Module, *args, **kwds):
    """Register a custom func or module in the computation graph.

    This should work will arbitrary functions and modules.

    In case of functions it might add a small delay to the call, because it is figuring out
    where the arguments should go. If this is unacceptable, please create a nn.Module from your func.

    All arguments, including Symbolic Tensors, should be passed after the ``func`` argument.
    They can be mixed and matched, even nested in lists, tuples and dictionaries.

    Convolution func example::

        inputs = Input(shape=(3, 32, 32))
        kernel = Input(shape=(16, 3, 3, 3), include_batch=False)
        bias = Input(shape=(16,), include_batch=False)
        output = add_function(F.conv2d, input=inputs, weight=k, bias=bias, padding=1)
    """
    if isinstance(func, nn.Module) and not kwds:
        return add_module_to_model(func, *args)

    extracted_symbols: List[SymbolicTensor] = []

    real_call_args = []
    real_call_kwds = {}

    navigation: List[List[Hashable]] = [[]]
    real_call_args = _replace_symbolic_with_value(args, extracted_symbols, navigation)
    real_call_kwds = _replace_symbolic_with_value(kwds, extracted_symbols, navigation)
    navigation.pop()

    assert len(extracted_symbols) > 0, "No Symbolic Tensors detected in the input!"
    assert all((isinstance(symbol, SymbolicTensor) for symbol in extracted_symbols))
    assert len(extracted_symbols) == len(navigation)

    def wrapper_function(*args):
        assert len(args) == len(navigation), f"Expected {len(navigation)} inputs, not {len(args)}!"
        for arg, navi in zip(args, navigation):
            obj = real_call_kwds if isinstance(navi[0], str) else real_call_args

            for idx in navi[:-1]:
                obj = obj[idx]

            obj[navi[-1]] = arg

        return func(*real_call_args, **real_call_kwds)

    module = useful_layers.NamedAnyOpLayer(op=wrapper_function, name=f"{func.__name__}({len(navigation)})")
    return extracted_symbols[0](module, *extracted_symbols[1:])



#  Copyright (c) 2022 Szymon Mikler

from __future__ import annotations

import logging
from typing import List, Set, Tuple

import torch
from torch import nn

from . import useful_layers


class SymbolicTensor:
    def __init__(
        self,
        value: torch.Tensor,
        parents: Tuple[SymbolicTensor, ...] = tuple(),
        depth: int = 0,
        layer: nn.Module | None = None,
        batch_size_known: bool = False,
    ):
        """Node in a SymbolicModel.

        This might represent input or intermediate values of the neural network.

        Parameters
        ----------
        value
            x object representing example of value that will flow through the node.
        parents
            Tuple of parent nodes - other SymbolicTensors.
        depth
            How deep the node is in the tree. It is defined as maximum of parents' depth plus 1.
        layer
            nn.Module object that transformed parents into this object.
        batch_size_known
            If False, the batch size will be replaced with ``None`` when displaying it.
        """
        self.v = value
        self.parents = parents
        self.children: List[SymbolicTensor] = []
        self.layer = layer
        self.depth = depth
        self._output = None
        self._parents_outputs: List[torch.Tensor] = []
        self.batch_size_known = batch_size_known

    @property
    def features(self) -> int:
        """Size of 1D data."""
        assert len(self.v.shape) == 2, "The data is not of [C,F] form!"
        return self.v.shape[1]

    @property
    def C(self) -> int:
        """Number of channels in Image data."""
        assert len(self.v.shape) == 4, "The data is not of [C,H,W] form!"
        return self.v.shape[1]

    @property
    def channels(self) -> int:
        """Same as ``.C``"""
        return self.C

    @property
    def H(self) -> int:
        """Height in Image data."""
        assert len(self.v.shape) == 4, "The data is not of [C,H,W] form!"
        return self.v.shape[2]

    @property
    def W(self) -> int:
        """Width in Image data."""
        assert len(self.v.shape) == 4, "The data is not of [C,H,W] form!"
        return self.v.shape[3]

    @property
    def HW(self) -> Tuple[int, int]:
        """Tuple of (height, width) in Image data."""
        return (self.H, self.W)

    @property
    def CHW(self) -> Tuple[int, int, int]:
        """Tuple of (channels, height, width) in Image data."""
        return (self.C, self.H, self.W)

    @property
    def HWC(self) -> Tuple[int, int, int]:
        """Tuple of (height, width, channels) in Image data."""
        return (self.H, self.W, self.C)

    @property
    def batch_size(self) -> int | None:
        """If known - batch size of the data. Else None."""
        if self.batch_size_known:
            return self.v.shape[0]
        else:
            return None

    @property
    def shape(self) -> Tuple[int | None, ...]:
        """Shape of the placeholder, including batch size."""
        if self.batch_size_known:
            return self.v.shape
        else:
            return (None, *self.v.shape[1:])

    @property
    def numel(self) -> int:
        """Number of the values in placeholder. If batch size is known, it is used too."""
        return self.v.shape.numel()

    def apply_module(self, layer: nn.Module, *others: SymbolicTensor) -> SymbolicTensor:
        """Register a new layer in the graph. Layer must be nn.Module."""
        assert all([isinstance(other, SymbolicTensor) for other in others])

        parents = (self, *others)
        new_depth = max(parent.depth for parent in parents) + 1
        new_output = layer.__call__(self.v, *(o.v for o in others))
        batch_size_known = all([parent.batch_size_known for parent in parents])

        new_layer_node = SymbolicTensor(
            value=new_output,
            parents=parents,
            layer=layer,
            depth=new_depth,
            batch_size_known=batch_size_known,
        )
        for parent in parents:
            parent.children.append(new_layer_node)
            logging.info(f"Added {new_layer_node} as child of {parent}")
        return new_layer_node

    def _get_all_nodes_above(self) -> Set[SymbolicTensor]:
        nodes_seen = {self}
        to_expand = [self]
        while to_expand:
            node = to_expand.pop()
            for parent in node.parents:
                if parent not in nodes_seen:
                    to_expand.append(parent)
                    nodes_seen.add(parent)
        return nodes_seen

    def _get_all_nodes_below(self) -> Set[SymbolicTensor]:
        nodes_seen = {self}
        to_expand = [self]
        while to_expand:
            node = to_expand.pop()
            for child in node.children:
                if child not in nodes_seen:
                    to_expand.append(child)
                    nodes_seen.add(child)
        return nodes_seen

    def _launch_input(self, x):
        self._output = x

    def _launch(self):
        self._output = self.layer(*(parent._output for parent in self.parents))

    def __call__(self, *args):
        return self.apply_module(*args)

    def __abs__(self):
        return self(useful_layers.AnyOpLayer(lambda x: abs(x)))

    def __add__(self, other):
        if isinstance(other, SymbolicTensor):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self(useful_layers.AddOpLayer(), other)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: x + other))

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, other):
        if isinstance(other, SymbolicTensor):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self(useful_layers.MulOpLayer(), other)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: x * other))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mod__(self, other):
        if isinstance(other, SymbolicTensor):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self(useful_layers.ModOpLayer(), other)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: x % other))

    def __rmod__(self, other):
        if isinstance(other, SymbolicTensor):
            return other.__mod__(self)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: other % x))

    def __neg__(self):
        return self(useful_layers.AnyOpLayer(op=lambda x: -x))

    def __pow__(self, other):
        if isinstance(other, SymbolicTensor):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self(useful_layers.AnyOpLayer(op=lambda x, y: x**y), other)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: x**other))

    def __rpow__(self, other):
        if isinstance(other, SymbolicTensor):
            return other.__pow__(self)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: other**x))

    def __sub__(self, other):
        if isinstance(other, SymbolicTensor):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self(useful_layers.SubOpLayer(), other)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: x - other))

    def __rsub__(self, other):
        if isinstance(other, SymbolicTensor):
            return other.__sub__(self)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: other - x))

    def __truediv__(self, other):
        if isinstance(other, SymbolicTensor):
            assert self.shape == other.shape, "Shapes do not match for the operation!"
            return self(useful_layers.AnyOpLayer(op=lambda x, y: x / y), other)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: x / other))

    def __rtruediv__(self, other):
        if isinstance(other, SymbolicTensor):
            return other.__truediv__(self)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: other / x))

    def __matmul__(self, other):
        if isinstance(other, SymbolicTensor):
            return self(useful_layers.MatmulOpLayer(), other)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: x @ other))

    def __rmatmul__(self, other):
        if isinstance(other, SymbolicTensor):
            return other.__matmul__(self)
        else:
            return self(useful_layers.AnyOpLayer(op=lambda x: other @ x))

    def __repr__(self):
        addr = f"SymbolicTensor at {hex(id(self))};"
        info = f"child of {len(self.parents)}; parent of {len(self.children)}"
        return "<" + addr + " " + info + ">"

    def __hash__(self):
        return id(self)


class Input(SymbolicTensor):
    def __init__(
        self,
        shape: Tuple | List | None = None,
        include_batch: bool = True,
        batch_shape: Tuple | List | None = None,
        dtype=torch.float32,
        min_value: float = 0.0,
        max_value: float = 1.0,
        custom_tensor: torch.Tensor | None = None,
    ):
        """Input to the SymbolicModel.

        It should be treated as a placeholder value that will be replaced with
        real data after the model is created.
        For calculation purposes, it can be treated as a normal numerical object,
        which means it can be added, subtracted, multiplied, taken absolute value of,
        etc.

        Parameters
        ----------
        shape
            Shape of the real data NOT including the batch dimension.
        include_batch
            If True and ``batch_shape`` was not given, batch_shape will not be added to ``shape``.
        batch_shape
            Shape of the real data including the batch dimension.
            Should be provided instead ``shape`` if cuda graphs will be used.
            If both ``shape`` and ``batch_shape`` are given, ``batch_shape`` has higher priority.
        dtype
            Dtype of the real data that will be the input of the network.
        min_value
            In rare cases, if real world data is very specific and some values
            cannot work with the model, this should be used to set the
            reasonable minimal value that the model will work on.
        max_value
            As above, but the maximal value.
        custom_tensor
            If needed, a specific tensor can be provided to serve as the SymbolicTensor's value.
            If this is the case, no shape or dtype is needed as they will be inferred from the tensor.
        """
        if custom_tensor is not None:
            self.was_batch_size_provided = True
            super().__init__(value=custom_tensor, batch_size_known=True)
            return

        if batch_shape is None and not include_batch:
            batch_shape = shape

        if batch_shape is not None:
            batch_size = batch_shape[0]
            shape = batch_shape[1:]
            batch_size_known = True
        elif shape is not None:
            # We use batch_size of 1 under the hood
            # but we don't tell it to the user
            batch_size = 1
            batch_size_known = False
        else:
            raise ValueError("Shape argument is required!")

        value = torch.rand(batch_size, *shape) * (max_value - min_value) + min_value
        value = value.to(dtype)
        super().__init__(value=value, batch_size_known=batch_size_known)



#  Copyright (c) 2022 Szymon Mikler

from __future__ import annotations

import copy
import logging
from types import MethodType
from typing import Any, Dict, List, Set, Tuple

import torch
from torch import nn

from . import code_generator, config
from .graph_algorithms import figure_out_nodes_between, topological_sort
from .symbolic_tensor import SymbolicTensor


class DetachedSymbolicModel(nn.Module):
    def __init__(self, names: List[str], layers: List[nn.Module], forward_src: str):
        """A tiny model detached from the SymbolicModel graph structure.

        It can live, even if the graph structure is removed!
        """
        super().__init__()
        self._execution_order_layers = []
        for name, layer in zip(names, layers):
            try:
                layer = copy.deepcopy(layer)
            except Exception as e:
                logging.error(f"Deepcopy of {layer} failed!")
                raise e
            self.add_module(name, layer)
            self._execution_order_layers.append(layer)

        self._generated_forward_source = forward_src

        scope = {"self": self}
        exec(self._generated_forward_source, {}, scope)
        self.forward = MethodType(scope["_generated_forward"], self)  # type: ignore


class SymbolicModel(nn.Module):
    def __init__(
        self,
        inputs: Tuple[SymbolicTensor, ...] | List[SymbolicTensor] | SymbolicTensor,
        outputs: Tuple[SymbolicTensor, ...] | List[SymbolicTensor] | SymbolicTensor,
        enable_cuda_graphs=False,
        enable_forward_codegen=True,
    ):
        """A PyTorch model that applies operations defined by the graph.

        All operations that changed ``inputs`` into ``outputs`` will be applied
        in the same order on the real data that will be fed into this model.

        Example::

            input1 = Input((10,))
            input2 = Input((10,))
            x = input1 + input2
            x = nn.Linear(x.features, 1)(x)
            model = SymbolicModel((input1, input2), x)

        Parameters
        ----------
        inputs
            A collection of SymbolicTensors that will begin the computations.
            For these nodes, you'll provide the input. So if you have mulitple inputs,
            be prepared to pass multiple inputs during training/inference.
        outputs
            A collection of SymbolicTensors that will end the computations.
            These nodes return your final computation result.
            So if you have mulitple outputs, SymbolicModel will return a tuple of tensors.
        enable_cuda_graphs
            If True, after the model creation, model will be converted to CUDA Graph.
            This requires CUDA capable device.
            CUDA Graphs are greatly speeding up the execution of some of the models.
            Not all models are compatible with CUDA Graphs. For example, if your model
            includes some non-deterministic behaviour, it likely won't work.
        """
        super().__init__()
        logging.info("Creating a SymbolicModel...")

        if isinstance(inputs, SymbolicTensor):
            inputs = (inputs,)
        assert all(isinstance(x, SymbolicTensor) for x in inputs)
        self.inputs: Tuple[SymbolicTensor, ...] = tuple(inputs)

        if isinstance(outputs, SymbolicTensor):
            outputs = (outputs,)
        assert all(isinstance(x, SymbolicTensor) for x in outputs)
        self.outputs: Tuple[SymbolicTensor, ...] = tuple(outputs)

        # Initialize helper variables
        self._node_to_layer_name: Dict[SymbolicTensor, str] = {}
        self._layer_name_to_node: Dict[str, SymbolicTensor] = {}
        self._execution_order_nodes: List[SymbolicTensor] = []
        self._execution_order_layers: List[nn.Module] = []
        self._figure_out_execution_order()

        self._enable_forward_codegen = enable_forward_codegen
        if self._enable_forward_codegen:
            self._replace_forward_with_codegen()

        self._enable_cuda_graphs = enable_cuda_graphs
        if self._enable_cuda_graphs:
            self._convert_to_cuda_graphs(self.inputs)

    def forward(self, *inputs: torch.Tensor) -> Any:
        """This function is executed by __call__. Do not use this directly, use __call__ instead.

        Warning!

        This function will be overwritten by `_generate_optimized_forward` if `enable_forward_codegen`
        is True. If this happened and you want to see your source, print `self._generated_forward_source`.
        """
        assert len(inputs) == len(self.inputs), "Number of inputs doesn't match!"
        for input_data, input_node in zip(inputs, self.inputs):
            input_node._launch_input(input_data)

        for node in self._execution_order_nodes:
            node._launch()

        if len(self.outputs) == 1:
            return self.outputs[0]._output
        else:
            return tuple(output_leaf._output for output_leaf in self.outputs)

    @property
    def input_shape(self):
        """Return shape of the input or in case of multiple inputs - a tuple of them."""
        if len(self.inputs) == 1:
            return self.inputs[0].shape
        else:
            return tuple(node.shape for node in self.inputs)

    @property
    def output_shape(self):
        """Return shape of the output or in case of multiple outputs - a atuple of them."""
        if len(self.outputs):
            return self.outputs[0].shape
        else:
            return tuple(node.shape for node in self.outputs)

    def add_output(self, node: SymbolicTensor):
        assert node not in self.inputs, "Node is an input!"
        assert node in self._execution_order_nodes, "Node is not in the graph!"

        self.outputs = (*self.outputs, node)
        if self._enable_forward_codegen:
            self._replace_forward_with_codegen()

    def detach_from_graph(self) -> DetachedSymbolicModel:
        if self._enable_cuda_graphs:
            logging.warning("This might fail after converting to CUDA Graphs!")

        names = [self._node_to_layer_name[node] for node in self._execution_order_nodes]
        forward_src = code_generator.generate_forward_with_loops(
            self.inputs,
            self.outputs,
            self._execution_order_nodes,
            min_loop_length=config.CODEGEN_MIN_LOOP_LENGTH,
        )
        return DetachedSymbolicModel(names, self._execution_order_layers, forward_src)

    def _replace_forward_with_codegen(self):
        self._generated_forward_source = code_generator.generate_forward_with_loops(
            self.inputs,
            self.outputs,
            self._execution_order_nodes,
            min_loop_length=config.CODEGEN_MIN_LOOP_LENGTH,
        )
        scope = {"self": self}
        exec(self._generated_forward_source, {}, scope)
        self.forward = MethodType(scope["_generated_forward"], self)

    def _convert_to_cuda_graphs(self, inputs: Tuple[SymbolicTensor, ...]):
        msg = (
            "CUDA Graphs can result in undefined behaviour! "
            "Please read https://pytorch.org/docs/stable/notes/cuda.html#constraints."
        )
        logging.warning(msg)
        assert torch.cuda.is_available(), "CUDA acceleration is not available!"
        for x in inputs:
            assert x.batch_size_known, "Must provide batch size for each input!"

        self.cuda()
        input_tensors = tuple(x.v.cuda() for x in inputs)
        torch.cuda.make_graphed_callables(self, sample_args=input_tensors)

    @property
    def _used_nodes(self) -> Set[SymbolicTensor]:
        """Return a set of all nodes used in this model."""
        return figure_out_nodes_between(self.inputs, self.outputs)

    def _figure_out_execution_order(self):
        # Exclude inputs, as we don't launch any layers in input nodes
        used_nodes_excl_inputs = self._used_nodes - set(self.inputs)
        self._execution_order_nodes = topological_sort(used_nodes_excl_inputs)
        self._execution_order_layers = [node.layer for node in self._execution_order_nodes]

        num_layers = len(self._execution_order_nodes)
        str_length = len(str(num_layers))
        for idx, node in enumerate(self._execution_order_nodes):
            name = f"module{str(idx).zfill(str_length)}_depth{str(node.depth).zfill(str_length)}"
            self._layer_name_to_node[name] = node
            self._node_to_layer_name[node] = name
            self.add_module(name=name, module=node.layer)

    def __deepcopy__(self, memo):
        """This copies a working Module, but the underlying graph structure is not copied!"""
        obj = self.detach_from_graph()
        memo[id(self)] = obj
        return obj



#  Copyright (c) 2022 Szymon Mikler

from . import graph_algorithms, model_tools
from .config import optimize_module_calls
from .functions_utility import add_to_model
from .symbolic_model import SymbolicModel
from .symbolic_tensor import Input, SymbolicTensor

__all__ = [
    "Input",
    "SymbolicModel",
    "SymbolicTensor",
    "add_to_model",
    "graph_algorithms",
    "model_tools",
    "optimize_module_calls",
]



# Copyright (c) 2022 Szymon Mikler

from __future__ import annotations

from typing import List, Tuple

from pytorch_symbolic.symbolic_tensor import SymbolicTensor


def generate_forward_basic(
    inputs: List[SymbolicTensor] | Tuple[SymbolicTensor, ...],
    outputs: List[SymbolicTensor] | Tuple[SymbolicTensor, ...],
    execution_order: List[SymbolicTensor] | Tuple[SymbolicTensor, ...],
):
    str_length = len(str(max(len(inputs), len(outputs), len(execution_order))))
    node_to_name = {}
    for idx, node in enumerate(inputs):
        node_to_name[node] = f"i{str(idx).zfill(str_length)}"
    for idx, node in enumerate(execution_order):
        node_to_name[node] = f"h{str(idx).zfill(str_length)}"
    for idx, node in enumerate(outputs):
        node_to_name[node] = f"o{str(idx).zfill(str_length)}"

    input_names = [node_to_name[node] for node in inputs]
    forward_definition = "def _generated_forward(self," + ",".join(input_names) + "):"
    code_lines = [forward_definition]

    TAB = " " * 4
    code_lines.append(TAB + "l = self._execution_order_layers")

    for exec_id, node in enumerate(execution_order):
        input_names = [node_to_name[node] for node in node.parents]
        output_name = node_to_name[node]
        code_line = TAB + output_name + f" = l[{exec_id}](" + ",".join(input_names) + ")"
        code_lines.append(code_line)

    return_line = TAB + "return " + ",".join(node_to_name[node] for node in outputs)
    code_lines.append(return_line)
    generated_forward = "\n".join(code_lines) + "\n"
    return generated_forward


def generate_forward_with_loops(
    inputs: List[SymbolicTensor] | Tuple[SymbolicTensor, ...],
    outputs: List[SymbolicTensor] | Tuple[SymbolicTensor, ...],
    execution_order: List[SymbolicTensor] | Tuple[SymbolicTensor, ...],
    min_loop_length: int | float = float("inf"),
):
    str_length = len(str(max(len(inputs), len(outputs), len(execution_order))))
    node_to_name = {}
    for idx, node in enumerate(inputs):
        node_to_name[node] = f"i{str(idx).zfill(str_length)}"
    for idx, node in enumerate(execution_order):
        node_to_name[node] = f"h{str(idx).zfill(str_length)}"
    for idx, node in enumerate(outputs):
        node_to_name[node] = f"o{str(idx).zfill(str_length)}"

    input_names = [node_to_name[node] for node in inputs]
    forward_definition = "def _generated_forward(self," + ",".join(input_names) + "):"
    code_lines = [forward_definition]

    TAB = " " * 4
    code_lines.append(TAB + "l = self._execution_order_layers")

    nodes_looped_over = set()
    all_nodes_used = {*inputs, *execution_order}
    # All parents must be in the graph. Otherwise, forward is impossible.
    parents = {node: node.parents for node in execution_order}
    # We only count children in the graph. Thus the intersection.
    children = {node: list(all_nodes_used.intersection(node.children)) for node in execution_order}

    for exec_id, node in enumerate(execution_order):
        if node in nodes_looped_over:
            continue

        input_names = [node_to_name[node] for node in node.parents]
        sequence = [node]
        last_node = node
        while (
            len(children[last_node]) == 1
            # this should never be false, but just in case we make sure the child is next in execution order
            and children[last_node][0] is execution_order[exec_id + len(sequence)]
            and len(parents[last_node]) == 1
            and len(parents[children[last_node][0]]) == 1
        ):
            last_node = children[last_node][0]
            sequence.append(last_node)

        if len(sequence) >= min_loop_length:
            output_name = node_to_name[sequence[-1]]
            code_lines.append(TAB + f"{output_name} = {input_names[0]}")
            code_lines.append(TAB + f"for layer in l[{exec_id}:{exec_id + len(sequence)}]:")
            code_lines.append(TAB + TAB + f"{output_name} = layer({output_name})")
            nodes_looped_over.update(sequence)
        else:
            output_name = node_to_name[node]
            code_line = TAB + output_name + f" = l[{exec_id}](" + ",".join(input_names) + ")"
            code_lines.append(code_line)

    code_lines.append(TAB + "return " + ",".join(node_to_name[node] for node in outputs))
    generated_forward = "\n".join(code_lines) + "\n"
    return generated_forward
